import logging
import os
import pickle
from enum import Enum
from uuid import UUID

import diskcache
from satcfdi.accounting import SatCFDI
from satcfdi.pacs import sat

from . import DATA_DIRECTORY
from .log_tools import print_yaml
from . import PPD

LIQUIDATED = 0
NOTIFIED = 2
STATUS_SAT = 3
FOLIO = 5

sat_manager = sat.SAT()

logger = logging.getLogger(__name__)


class LocalDB(diskcache.Cache):
    def __init__(self):
        super().__init__(directory=os.path.join(DATA_DIRECTORY, 'cache'))

    def folio(self) -> int:
        return self.get(FOLIO, 1)

    def folio_set(self, value: int):
        self[FOLIO] = value

    def liquidated(self, uuid: UUID):
        return self.get((LIQUIDATED, uuid))

    def liquidated_set(self, uuid: UUID, value: bool):
        self[(LIQUIDATED, uuid)] = value

    def notified(self, uuid: UUID):
        return self.get((NOTIFIED, uuid))

    def notified_set(self, uuid: UUID, value: bool):
        self[(NOTIFIED, uuid)] = value

    def status_sat(self, uuid: UUID):
        return self.get((STATUS_SAT, uuid), {})

    def status_sat_set(self, uuid: UUID, value: dict):
        if value:
            self[(STATUS_SAT, uuid)] = value
        else:
            try:
                del self[(STATUS_SAT, uuid)]
            except KeyError:
                pass


class LiquidatedState(Enum):
    NONE = 1
    YES = 2
    NO = 3
    IGNORED = 4
    CANCELLED = 5

    def __str__(self):
        if self.name == "NONE":
            return ""
        if self.name == "IGNORED":
            return "Ignorada"
        if self.name == "YES":
            return "Si"
        if self.name == "NO":
            return "No"
        if self.name == "CANCELLED":
            return "Cancelada"


class LocalDBSatCFDI(LocalDB):
    def __init__(self, enviar_a_partir, pagar_a_partir):
        super().__init__()
        self.enviar_a_partir = enviar_a_partir
        self.pagar_a_partir = pagar_a_partir

    def notified(self, cfdi: SatCFDI):
        v = super().notified(cfdi.uuid)
        if v is None and cfdi["Fecha"] < self.enviar_a_partir:
            return True
        return v

    def notified_flip(self, cfdi: SatCFDI):
        v = not self.notified(cfdi)
        self.notified_set(cfdi.uuid, v)
        return v

    def liquidated(self, cfdi: SatCFDI):
        v = super().liquidated(cfdi.uuid)
        if v is None and cfdi["Fecha"] < self.pagar_a_partir[cfdi["MetodoPago"]]:
            return True
        return v

    def liquidated_flip(self, cfdi: SatCFDI):
        v = not self.liquidated(cfdi)
        self.liquidated_set(cfdi.uuid, v)
        return v

    def status_sat(self, cfdi: SatCFDI, update=False):
        if update:
            res = sat_manager.status(cfdi)
            if res["ValidacionEFOS"] == "200":
                self.status_sat_set(cfdi.uuid, res)
            return res
        else:
            return super().status_sat(cfdi.uuid)

    def liquidated_state(self, cfdi: SatCFDI):
        if cfdi.estatus == '0':
            return LiquidatedState.CANCELLED

        if cfdi["TipoDeComprobante"] != "I":
            return LiquidatedState.NONE

        mpago = cfdi["MetodoPago"]
        if cfdi['Total'] == 0 or (mpago == PPD and cfdi.saldo_pendiente == 0):
            return LiquidatedState.YES

        if self.liquidated(cfdi):
            if mpago == PPD:
                return LiquidatedState.IGNORED
            return LiquidatedState.YES

        return LiquidatedState.NO

    def describe(self, cfdi: SatCFDI):
        print_yaml({
            'saldada': self.liquidated(cfdi),
            'enviada': self.notified(cfdi),
            'status_sat': self.status_sat(cfdi)
        })


def save_data(file, data):
    with open(os.path.join(DATA_DIRECTORY, file), 'wb') as f:
        pickle.dump(data, f)


def load_data(file, default=None):
    try:
        with open(os.path.join(DATA_DIRECTORY, file), 'rb') as f:
            return pickle.load(f)
    except FileNotFoundError:
        return default



import yaml
from satcfdi import Code


class NoAliasDumper(yaml.SafeDumper):
    def ignore_aliases(self, data):
        return True


def print_yaml(data):
    print(
        yaml.dump(data, Dumper=NoAliasDumper, allow_unicode=True, width=1280, sort_keys=False)
    )


def header_line(text):
    ln = (150 - len(text)) // 2
    return ("=" * ln) + " " + text + " " + ("=" * ln)


def cfdi_header(cfdi):
    receptor = Code(cfdi['Receptor']['Rfc'], cfdi['Receptor']['Nombre'])
    return f"{cfdi.name} - {cfdi.uuid} {receptor}"



import base64
import io
import itertools
import logging
import os
from datetime import timedelta, date
from zipfile import ZipFile

from PySimpleGUI import POPUP_BUTTONS_OK_CANCEL, PySimpleGUI
from satcfdi import DatePeriod, csf
from satcfdi.accounting import EmailManager
from satcfdi.create.cfd import cfdi40
from satcfdi.exceptions import ResponseError
from satcfdi.pacs import Accept
from satcfdi.pacs.sat import SAT, TipoDescargaMasivaTerceros, EstadoSolicitud
from satcfdi.printer import Representable
# noinspection PyUnresolvedReferences
from satcfdi.transform.catalog import CATALOGS

from . import __version__, PPD, PUE, TEMP_DIRECTORY
from .client_validation import validar_client
from .environments import facturacion_environment
from .file_data_managers import ClientsManager, FacturasManager
from .formatting_functions.common import fecha, pesos, porcentaje
from .gui_functions import generate_ingresos, pago_factura, exportar_facturas, archivos_filename, mf_pago_fmt, archivos_folder, period_desc, ajustes_directory, find_ajustes, \
    format_concepto_desc, generate_pdf_template, periodicidad_desc, parse_fecha_pago, parse_importe_pago, preview_cfdis
from .layout import make_layout, ActionButtonManager
from .localdb import LocalDBSatCFDI, LiquidatedState
from .log_tools import cfdi_header, header_line, print_yaml
from .mycfdi import get_all_cfdi, MyCFDI, move_to_folder
from .utils import random_string, parse_date_period, load_certificate, to_int, cert_info, add_month, clear_directory, find_best_match, months_between

logging.getLogger("weasyprint").setLevel(logging.ERROR)
logging.getLogger("fontTools").setLevel(logging.ERROR)

logger = logging.getLogger(__name__)


def open_launch_window():
    layout = [[PySimpleGUI.Text("New Window", key="new")]]
    window = PySimpleGUI.Window("Launch Window", layout, modal=True, size=(300, 300))
    window.read(timeout=1000)

    return window


class FacturacionGUI:
    def __init__(self, config):
        os.makedirs(TEMP_DIRECTORY, exist_ok=True)
        self.email_manager = EmailManager(
            **config['email']
        )
        pac = config['pac']
        pac_module, pac_class = pac['type'].split(".")
        mod = __import__(f"satcfdi.pacs.{pac_module}", fromlist=[pac_class])
        self.pac_service = getattr(mod, pac_class)(
            **pac['args']
        )
        self.serie = config['serie']
        self.csd_signer = load_certificate(config.get('csd')) if 'csd' in config else None
        self.fiel_signer = load_certificate(config.get('fiel')) if 'fiel' in config else None

        self.sat_service = SAT(signer=self.fiel_signer)
        self.rfc_prediales = config['rfc_prediales']

        self.window = PySimpleGUI.Window(
            f"Facturación Mensual CFDI 4.0 {self.csd_signer.rfc}",
            make_layout(bool(self.fiel_signer)),
            size=(1280, 800),
            resizable=True,
            font=("Courier New", 10, "bold"),
        )

        self.all_invoices = None
        self.local_db = LocalDBSatCFDI(
            enviar_a_partir=config['enviar_a_partir'],
            pagar_a_partir=config['pagar_a_partir']
        )
        # noinspection PyTypeChecker
        self.selected_satcfdi = None  # type: MyCFDI

        self.action_button_manager = ActionButtonManager(self.window["crear_facturas"])
        self.console = self.window["console"]

        MyCFDI.local_db = self.local_db

    def run(self):
        self.window.finalize()

        self.set_folio()
        self.window['serie'].update(self.serie)

        self.window['periodo'].bind("<Return>", "_enter")
        self.window['importe_pago'].bind("<Return>", "_enter")
        self.window['fecha_pago'].bind("<Return>", "_enter")
        self.window['forma_pago'].bind("<Return>", "_enter")
        self.window['emitidas_search'].bind("<Return>", "_enter")

        self.window['facturas_table'].bind('<Double-Button-1>', '_double_click')
        self.window['clientes_table'].bind('<Double-Button-1>', '_double_click')
        self.window['emitidas_table'].bind('<Double-Button-1>', '_double_click')
        self.window['ajustes_table'].bind('<Double-Button-1>', '_double_click')

        # Add logging to the window
        h = logging.StreamHandler(self.window['console'])
        h.setLevel(logging.INFO)
        logging.root.addHandler(h)

        self.main_loop()
        self.window.close()

    def initial_screen(self, emisor_cif):
        self.header("ACERCA DE")
        print_yaml({
            "version": __version__.__version__,
            "facturacion": "CFDI 4.0",
            "emisor": emisor_cif,
            "pac_service": {
                "Type": type(self.pac_service).__name__,
                "Rfc": self.pac_service.RFC,
                "Environment": str(self.pac_service.environment)
            },
            "fiel": cert_info(self.fiel_signer),
            "csd": cert_info(self.csd_signer),
        })

    def get_all_invoices(self):
        if self.all_invoices:
            return self.all_invoices
        self.all_invoices = get_all_cfdi()
        return self.all_invoices

    def generate_invoice(self, invoice):
        ref_id = random_string()

        # Add Serie and Folio and signature
        invoice['Serie'] = self.serie
        folio = self.local_db.folio()
        invoice['Folio'] = str(folio)
        cfdi40.Comprobante.sign(invoice, self.csd_signer)

        attempts = 3
        for i in range(attempts):
            if i:
                print(f'Intentando de nuevo... Intento {i + 1} de {attempts}')
                self._read(timeout=2000 * i)

            try:
                res = self.pac_service.stamp(
                    cfdi=invoice,
                    accept=Accept.XML_PDF,
                    ref_id=ref_id
                )
            except Exception as ex:
                logger.error(f"Error Generando: {invoice.get('Serie')}{invoice.get('Folio')} {invoice['Receptor']['Rfc']}")
                if isinstance(ex, ResponseError):
                    logger.error(f"Status Code: {ex.response.status_code}")
                    logger.error(f"Response: {ex.response.text}")
                continue

            self.set_folio(folio + 1)
            return move_to_folder(res.xml, pdf_data=res.pdf)

    def set_folio(self, folio: int = None):
        if folio:
            self.local_db.folio_set(folio)
        self.window['folio'].update(folio or self.local_db.folio())

    def recupera_comprobantes(self, id_solicitud):
        response = self.sat_service.recover_comprobante_status(
            id_solicitud=id_solicitud
        )
        print_yaml(response)
        self._read()
        if response["EstadoSolicitud"] == EstadoSolicitud.Terminada:
            for id_paquete in response['IdsPaquetes']:
                response, paquete = self.sat_service.recover_comprobante_download(
                    id_paquete=id_paquete
                )
                print_yaml(response)
                self._read()
                yield id_paquete, base64.b64decode(paquete) if paquete else None

    def unzip_cfdi(self, file):
        with ZipFile(file, "r") as zf:
            for fileinfo in zf.infolist():
                xml_data = zf.read(fileinfo)
                move_to_folder(xml_data, pdf_data=None)
                self._read()

    def _read(self, timeout=0):
        event, values = self.window.read(timeout=timeout)
        if event in ("Exit", PySimpleGUI.WIN_CLOSED):
            exit(0)

    def action_button(self, action_name, action_items):
        if action_name == 'facturas':
            self.all_invoices = None
            for invoice in action_items:
                cfdi = self.generate_invoice(invoice=invoice)
                if cfdi is None:
                    break
                print_yaml({
                    "FacturaGenerada": cfdi_header(cfdi),
                })
                self._read()
        elif action_name == 'correos':
            clients = ClientsManager()
            emisor_cif = clients[self.csd_signer.rfc]
            with self.email_manager.sender as s:
                for receptor, notify_invoices, pendientes_meses_anteriores in action_items:
                    def attachments():
                        for ni in notify_invoices:
                            yield ni.filename + ".xml"
                            yield ni.filename + ".pdf"

                    subject = f"Comprobantes Fiscales {receptor['RazonSocial']} - {receptor['Rfc']}"

                    s.send_email(
                        subject=subject,
                        to_addrs=receptor["Email"],
                        html=facturacion_environment.get_template('mail_facturas_template.html').render(
                            facturas=notify_invoices,
                            pendientes_meses_anteriores=pendientes_meses_anteriores,
                            emisor=emisor_cif,
                        ),
                        file_attachments=attachments()
                    )
                    for r in notify_invoices:
                        self.local_db.notified_set(r.uuid, True)
                    print_yaml({
                        "correo": subject,
                        "para": receptor["Email"]
                    })
                    self._read()
        elif action_name == 'clientes':
            for client in action_items:
                print(f"Validando: {client['Rfc']}")
                validar_client(client)
                self._read()
        else:
            raise ValueError(f"Invalid action: {action_name}")
        print("FIN")

    def set_selected_satcfdi(self, factura):
        i = factura
        self.selected_satcfdi = i
        if i:
            estado = self.local_db.status_sat(i).get('Estado', 'Vigente')
            self.window["status_sat"].update(
                estado.center(10),
                disabled=False,
                button_color="red4" if estado != "Vigente" else "green",
            )
        else:
            self.window["status_sat"].update(
                "".ljust(10), disabled=True, button_color=PySimpleGUI.theme_background_color()
            )

        # Email
        is_enviable = i \
                      and i["Emisor"]["Rfc"] == self.csd_signer.rfc \
                      and i.estatus == "1"
        if is_enviable:
            self.window["email_notificada"].update(
                "Enviada".center(10) if self.local_db.notified(i) else "Por Enviar",
                disabled=False,
                button_color="green" if self.local_db.notified(i) else "red4",
            )
        else:
            self.window["email_notificada"].update(
                "".ljust(10), disabled=True, button_color=PySimpleGUI.theme_background_color()
            )

        # Pendiente de Pago
        is_pendientable = i \
                          and i["Emisor"]["Rfc"] == self.csd_signer.rfc \
                          and i["TipoDeComprobante"] == "I" \
                          and i.estatus == "1" \
                          and (i["MetodoPago"] == PUE or i.saldo_pendiente) \
                          and i["Total"]
        if is_pendientable:
            self.window["pendiente_pago"].update(
                ("Pagada".center(10) if i["MetodoPago"] == PUE else "Ignorada".center(10)) if self.local_db.liquidated(i) else "Por Pagar".center(10),
                disabled=False,
                button_color="green" if self.local_db.liquidated(i) else "red4",
            )
        else:
            self.window["pendiente_pago"].update(
                "".ljust(10), disabled=True, button_color=PySimpleGUI.theme_background_color()
            )

        # PPD
        is_ppd_active = i \
                        and i["Emisor"]["Rfc"] == self.csd_signer.rfc \
                        and i.get("MetodoPago") == PPD \
                        and i.estatus == "1" \
                        and i.saldo_pendiente
        is_ppd_active = bool(is_ppd_active)
        self.window["ppd_action_items"].update(visible=is_ppd_active)
        self.window["importe_pago"].update("")

    def header(self, name):
        self.console.update("")
        print(header_line(name))

    def facturas_search(self, search_text):
        search_text = search_text.strip().upper()
        if len(search_text) < 3:
            self.window["emitidas_text"].update("El texto de búsqueda debe tener al menos 3 caracteres")
            return

        self.window["emitidas_text"].update(search_text)

        def fact_iter():
            for i in self.get_all_invoices().values():
                if i["Emisor"]["Rfc"] == self.csd_signer.rfc \
                        and (
                        i.name == search_text
                        or i["Receptor"]["Rfc"] == search_text
                        or search_text in i["Receptor"].get("Nombre", "")
                ):
                    yield i

        cfdis = list(fact_iter())
        self.emitidas_show(cfdis)

    def facturas_pendientes(self):
        self.window["emitidas_text"].update("Facturas Pendientes de Pago")

        def fact_iter():
            for i in self.get_all_invoices().values():
                if i["Emisor"]["Rfc"] == self.csd_signer.rfc \
                        and self.local_db.liquidated_state(i) == LiquidatedState.NO:
                    yield i

        cfdis = list(fact_iter())
        self.emitidas_show(cfdis)

    def facturas_emitidas(self, dp):
        self.window["emitidas_text"].update("Facturas Emitidas en " + str(dp))

        def fact_iter():
            for i in self.get_all_invoices().values():
                if i["Emisor"]["Rfc"] == self.csd_signer.rfc \
                        and i["Fecha"] == dp:
                    yield i

        cfdis = list(fact_iter())
        self.emitidas_show(cfdis)

    def emitidas_show(self, cfdis):
        rows = [
            [
                i,
                cfdi['Receptor'].get('Nombre', '')[0:36],
                cfdi['Receptor']['Rfc'],
                cfdi.name,
                cfdi["Fecha"].strftime("%Y-%m-%d"),
                cfdi["Total"],
                self.local_db.liquidated_state(cfdi),
                mf_pago_fmt(cfdi)
            ]
            for i, cfdi in enumerate(cfdis, start=1)
        ]
        self.window['emitidas_table'].update(
            values=rows,
        )
        self.window['emitidas_table'].metadata = cfdis

    def main_loop(self):
        while True:
            event, values = self.window.read()
            try:
                if event in ("Exit", PySimpleGUI.WIN_CLOSED):
                    return

                try:
                    dp = parse_date_period(values["periodo"])
                except ValueError:
                    dp = None

                if event in ("periodo_enter", "refresh_facturas", "refresh_ajustes", "refresh_clientes", "refresh_correos"):
                    event = 'main_tab_group'

                match event:
                    case "folio":
                        self.set_folio(to_int(values["folio"]))

                    case "about":
                        clients = ClientsManager()
                        self.initial_screen(clients[self.csd_signer.rfc])

                    case "recuperar_emitidas" | "recuperar_recibidas":
                        fecha_final = date.today()
                        fecha_inicial = fecha_final - timedelta(days=int(values["recuperar_dias"]))
                        id_solicitud = self.local_db.get(event)

                        if not id_solicitud:
                            self._read()
                            response = self.sat_service.recover_comprobante_request(
                                fecha_inicial=fecha_inicial,
                                fecha_final=fecha_final,
                                rfc_receptor=self.sat_service.signer.rfc if "recuperar_recibidas" == event else None,
                                rfc_emisor=self.sat_service.signer.rfc if "recuperar_emitidas" == event else None,
                                tipo_solicitud=TipoDescargaMasivaTerceros.CFDI,
                            )
                            print_yaml(response)
                            self.local_db[event] = response['IdSolicitud']
                            print("Nueva Solicitud Creada")
                        else:
                            print_yaml({
                                'IdSolicitud': id_solicitud
                            })
                            self._read()
                            for paquete_id, data in self.recupera_comprobantes(id_solicitud):
                                if data:
                                    self.all_invoices = None
                                    with io.BytesIO(data) as b:
                                        self.unzip_cfdi(b)
                                del self.local_db[event]
                            print("FIN")

                    case "emitidas_search_enter":
                        self.action_button_manager.clear()
                        self.facturas_search(values["emitidas_search"])

                    case 'main_tab_group':
                        if values['main_tab_group'] == 'console_tab':
                            self.action_button_manager.clear()

                        if values['main_tab_group'] == 'recuperar_tab':
                            self.action_button_manager.clear()

                        if values['main_tab_group'] == 'facturas_tab':
                            self.window['facturas_table'].update(values=[])

                            if dp is None or dp.month is None:
                                self.window['preparar_facturas_text'].update("Periodo no válido")
                                continue

                            self.window['preparar_facturas_text'].update(f"{period_desc(dp)}")
                            self.console.update("")
                            cfdis = generate_ingresos(
                                clients=ClientsManager(),
                                facturas=FacturasManager(dp)["Facturas"],
                                dp=dp,
                                emisor_rfc=self.csd_signer.rfc
                            )
                            if cfdis is None:
                                self.window['console_tab'].select()
                                continue

                            rows = [[
                                i,
                                cfdi['Emisor']['RegimenFiscal'].code,
                                cfdi['Receptor']['Nombre'][0:36],
                                cfdi['Receptor']['Rfc'],
                                mf_pago_fmt(cfdi),
                                cfdi['SubTotal'],
                                cfdi['Total'],
                            ]
                                for i, cfdi in enumerate(cfdis, start=1)
                            ]
                            self.window['facturas_table'].update(
                                values=rows,
                            )
                            self.window['facturas_table'].metadata = cfdis
                            self.action_button_manager.set_items('facturas', cfdis)

                        if values['main_tab_group'] == 'clients_tab':
                            self.window['clientes_table'].update(values=[])
                            clients = list(ClientsManager().values())
                            facturas = FacturasManager(dp)["Facturas"]
                            rows = [
                                [
                                    i,
                                    client["RazonSocial"][:36],
                                    client["Rfc"],
                                    client["RegimenFiscal"].code,
                                    client["CodigoPostal"],
                                    sum(1 for f in facturas if f["Receptor"] == client["Rfc"]) or "",
                                ]
                                for i, client in enumerate(clients, start=1)
                            ]
                            self.window['clientes_table'].update(
                                values=rows,
                            )
                            self.window['clientes_table'].metadata = clients
                            self.action_button_manager.set_items('clientes', clients)

                        if values['main_tab_group'] == 'emitidas_tab':
                            self.action_button_manager.clear()
                            self.facturas_pendientes()

                        if values['main_tab_group'] == 'correos_tab':
                            now = date.today()
                            dp_now = DatePeriod(now.year, now.month)
                            clients = ClientsManager()
                            a_invoices = self.get_all_invoices()

                            cfdi_correos = []
                            for receptor_rfc, notify_invoices in itertools.groupby(
                                    sorted(
                                        (i for i in a_invoices.values() if i.estatus == "1" and not self.local_db.notified(i)),
                                        key=lambda r: r["Receptor"]["Rfc"]
                                    ),
                                    lambda r: r["Receptor"]["Rfc"]
                            ):
                                notify_invoices = list(notify_invoices)

                                def fac_pen_iter():
                                    for i in self.get_all_invoices().values():
                                        if i["Emisor"]["Rfc"] == self.csd_signer.rfc \
                                                and self.local_db.liquidated_state(i) == LiquidatedState.NO \
                                                and i["Fecha"] < dp_now \
                                                and i["Receptor"]["Rfc"] == receptor_rfc \
                                                and i not in notify_invoices:
                                            yield i

                                fac_pen = sorted(
                                    fac_pen_iter(),
                                    key=lambda r: r["Fecha"]
                                )
                                receptor = clients[receptor_rfc]
                                cfdi_correos.append((receptor, notify_invoices, fac_pen))

                            rows = [
                                [
                                    i,
                                    receptor["RazonSocial"][0:36],
                                    receptor["Rfc"],
                                    ",".join(n.name for n in notify_invoices),
                                    ",".join(n.name for n in facturas_pendientes)
                                ] for i, (receptor, notify_invoices, facturas_pendientes) in enumerate(cfdi_correos, start=1)
                            ]
                            self.window['correos_table'].update(
                                values=rows,
                            )
                            self.window['correos_table'].metadata = cfdi_correos
                            self.action_button_manager.set_items('correos', cfdi_correos)

                        if values['main_tab_group'] == 'ajustes_tab':
                            if dp is None or dp.month is None:
                                self.window['preparar_ajustes_text'].update("Periodo no válido")
                                self.window['ajustes_table'].update(values=[])
                                continue

                            dp_effective = add_month(dp, 1)
                            self.window['preparar_ajustes_text'].update(f"Ajustes Efectivos Al: {period_desc(dp_effective)}")

                            # clear directory
                            ajustes_dir = ajustes_directory(DatePeriod(dp.year, dp.month))
                            clear_directory(ajustes_dir)
                            emisor_rfc = self.csd_signer.rfc

                            clients = ClientsManager()
                            facturas = FacturasManager(None)["Facturas"]

                            def ajustes_iter():
                                for i, (rfc, concepto) in enumerate(find_ajustes(facturas, dp_effective.month)):
                                    receptor = clients[rfc]  # type: dict
                                    valor_unitario_raw = concepto["ValorUnitario"]

                                    if isinstance(valor_unitario_raw, dict):
                                        vud, vu = find_best_match(valor_unitario_raw, dp)
                                        vund, vun = find_best_match(valor_unitario_raw, dp_effective)
                                        meses = months_between(vund, vud)
                                        ajuste_porcentaje = (vun / vu - 1)
                                    else:
                                        vu = valor_unitario_raw
                                        vun = None
                                        meses = None
                                        ajuste_porcentaje = None

                                    concepto = format_concepto_desc(concepto, periodo="INMUEBLE")
                                    file_name = os.path.join(ajustes_dir, f'AjusteRenta_{rfc}_{i}.pdf')

                                    data = {
                                        "receptor": receptor,
                                        "emisor": clients[emisor_rfc],
                                        "concepto": concepto,
                                        "valor_unitario": pesos(vu),
                                        "valor_unitario_nuevo": pesos(vun) if vun else "",
                                        "ajuste_porcentaje": porcentaje(ajuste_porcentaje, 2) if ajuste_porcentaje is not None else "",
                                        "ajuste_periodo": f"{meses} MESES",
                                        "efectivo_periodo_desc": periodicidad_desc(dp_effective, concepto['_periodo_mes_ajuste'], concepto.get('_desfase_mes')),
                                        "periodo": concepto['_periodo_mes_ajuste'].split('.')[0].upper(),
                                        "fecha_hoy": fecha(date.today()),
                                        'file_name': file_name
                                    }
                                    if ajuste_porcentaje is not None:
                                        res = generate_pdf_template(
                                            template_name='incremento_template.md',
                                            fields=data
                                        )
                                        with open(file_name, 'wb') as f:
                                            f.write(res)
                                    yield data

                            ajustes_iter = list(ajustes_iter())
                            rows = [
                                [
                                    i,
                                    ajuste["receptor"]["RazonSocial"][:36],
                                    ajuste["receptor"]["Rfc"],
                                    ajuste["valor_unitario"].split(' ')[0],
                                    ajuste["valor_unitario_nuevo"].split(' ')[0],
                                    ajuste["ajuste_porcentaje"].split(' ')[0],
                                    ajuste["periodo"],
                                    ajuste["ajuste_periodo"],
                                    ajuste["efectivo_periodo_desc"]
                                ]
                                for i, ajuste in enumerate(ajustes_iter, start=1)
                            ]
                            self.window['ajustes_table'].update(
                                values=rows,
                            )
                            self.window['ajustes_table'].metadata = ajustes_iter
                            self.action_button_manager.set_items('ajustes', ajustes_iter)

                    case "facturas_pendientes":
                        self.facturas_pendientes()

                    case "facturas_emitidas":
                        self.facturas_emitidas(dp)

                    case 'facturas_table_double_click':
                        items = self.window["facturas_table"].metadata
                        selection = values["facturas_table"]
                        s_items = [items[i] for i in selection] if selection else items
                        preview_cfdis(s_items)

                    case 'clientes_table_double_click':
                        items = self.window["clientes_table"].metadata
                        selection = values["clientes_table"]
                        s_items = [items[i] for i in selection] if selection else items

                        if s_items:
                            client = s_items[0]
                            url = csf.url(rfc=client["Rfc"], id_cif=client["IdCIF"])
                            os.startfile(url)

                    case 'emitidas_table_double_click':
                        items = self.window["emitidas_table"].metadata
                        selection = values["emitidas_table"]
                        s_items = [items[i] for i in selection] if selection else items

                        if s_items:
                            cfdi = s_items[0]  # type: MyCFDI
                            os.startfile(
                                os.path.abspath(cfdi.filename + ".pdf")
                            )

                    case 'ajustes_table_double_click':
                        items = self.window["ajustes_table"].metadata
                        selection = values["ajustes_table"]
                        s_items = [items[i] for i in selection] if selection else items

                        if s_items:
                            ajuste = s_items[0]
                            os.startfile(
                                os.path.abspath(ajuste['file_name'])
                            )

                    case "facturas_table" | "clientes_table" | "correos_table" | "ajustes_table" | "emitidas_table":
                        items = self.window[event].metadata
                        selection = values[event]
                        if event == "emitidas_table":
                            if selection:
                                self.set_selected_satcfdi(
                                    items[selection[0]]
                                )
                            else:
                                self.set_selected_satcfdi([])
                        else:
                            s_items = [items[i] for i in selection] if selection else items
                            self.action_button_manager.set_items(event.split("_")[0], s_items)

                    case "prepare_pago" | "importe_pago_enter" | "fecha_pago_enter" | "forma_pago_enter" | "ver_html_pago":
                        if i := self.selected_satcfdi:
                            fecha_pago = parse_fecha_pago(values["fecha_pago"])
                            importe_pago = parse_importe_pago(values["importe_pago"])
                            importe_pago = importe_pago or i.saldo_pendiente
                            self.window["importe_pago"].update(importe_pago)

                            cfdi = pago_factura(
                                factura_pagar=i,
                                fecha_pago=fecha_pago,
                                forma_pago=values["forma_pago"],
                                importe_pago=importe_pago,
                            )
                            self.action_button_manager.set_items('facturas', [cfdi])
                            if event == "ver_html_pago":
                                preview_cfdis([cfdi])

                    case "status_sat":
                        if i := self.selected_satcfdi:
                            self.local_db.status_sat(i, update=True)
                            self.set_selected_satcfdi(i)

                    case "pendiente_pago":
                        if i := self.selected_satcfdi:
                            self.local_db.liquidated_flip(i)
                            self.set_selected_satcfdi(i)

                    case "email_notificada":
                        if i := self.selected_satcfdi:
                            self.local_db.notified_flip(i)
                            self.set_selected_satcfdi(i)

                    case "crear_facturas":
                        action_text = self.action_button_manager.text()
                        res = PySimpleGUI.popup(
                            f"{action_text}?",
                            title="Confirmar",
                            button_type=POPUP_BUTTONS_OK_CANCEL,
                        )
                        if res == "OK":
                            self.header(action_text.upper())
                            self.window['console_tab'].select()
                            self._read()
                            self.action_button(
                                action_name=self.action_button_manager.name,
                                action_items=self.action_button_manager.items
                            )
                            self.action_button_manager.clear()

                    case "ver_excel":
                        clients = ClientsManager()
                        emisor_cif = clients[self.csd_signer.rfc]
                        if archivo_excel := exportar_facturas(
                                self.get_all_invoices(),
                                dp,
                                emisor_cif,
                                self.rfc_prediales
                        ):
                            os.startfile(
                                os.path.abspath(archivo_excel)
                            )
                        else:
                            raise Exception("No se pudo crear el archivo, cierra el archivo si se tiene abierto")

                    case "ver_html":
                        def fact_iter():
                            for i in self.get_all_invoices().values():
                                if i["Fecha"] == dp:
                                    yield i

                        if cfdis := list(fact_iter()):
                            outfile = archivos_filename(dp, ext="html")
                            Representable.html_write_all(
                                objs=cfdis,
                                target=outfile,
                            )
                            os.startfile(
                                os.path.abspath(outfile)
                            )
                        else:
                            raise Exception("No hay facturas para el periodo seleccionado")

                    case "ver_carpeta":
                        directory = archivos_folder(dp)
                        os.startfile(
                            os.path.abspath(directory)
                        )

                    case "ver_config":
                        os.startfile(
                            os.path.abspath(".")
                        )

                    case "editar_clientes":
                        os.startfile(
                            os.path.abspath("clientes.yaml")
                        )

                    case "editar_facturas":
                        os.startfile(
                            os.path.abspath("facturas.yaml")
                        )

                    case "sat_status_todas":
                        def fact_iter():
                            for i in self.get_all_invoices().values():
                                if i["Fecha"] == dp:
                                    yield i

                        for cfdi in fact_iter():
                            print(f"Estado SAT: {cfdi_header(cfdi)}")
                            self._read()
                            estado = self.local_db.status_sat(cfdi, update=True)
                            print_yaml(estado)
                        print("FIN")

                    case "periodo" | "inicio" | "final" | "fecha_pago" | "forma_pago" | "importe_pago" | ' ':
                        pass

                    case _:
                        logger.error(f"Unknown event '{event}'")

            except Exception as ex:
                logger.exception(header_line("ERROR"))
                if values['main_tab_group'] != 'console_tab':
                    PySimpleGUI.Popup(
                        ex,
                        no_titlebar=True,
                        grab_anywhere=True,
                        any_key_closes=True,
                        background_color="red4",
                    )



import logging
import os
from datetime import datetime, date
from decimal import Decimal
from decimal import InvalidOperation

import xlsxwriter
from babel.dates import format_date
from markdown2 import markdown
from satcfdi import DatePeriod
from satcfdi.accounting import filter_invoices_iter, filter_payments_iter, invoices_export, payments_export
from satcfdi.accounting.process import payments_groupby_receptor, payments_retentions_export
from satcfdi.create.cfd import cfdi40
from satcfdi.create.cfd.cfdi40 import Comprobante, PagoComprobante
from satcfdi.pacs import sat
from satcfdi.printer import Representable
# noinspection PyUnresolvedReferences
from satcfdi.transform.catalog import CATALOGS
from weasyprint import HTML, CSS
from xlsxwriter.exceptions import FileCreateError

from . import SOURCE_DIRECTORY, ARCHIVOS_DIRECTORY, TEMP_DIRECTORY
from .environments import facturacion_environment
from .utils import add_month

logger = logging.getLogger(__name__)
logger.level = logging.INFO

sat_manager = sat.SAT()

PERIODICIDAD = {
    "Mensual": 1,
    "Bimestral": 2,
    "Trimestral": 3,
    "Cuatrimestral": 4,
    "Semestral": 6,
    "Anual": 12
}


def create_cfdi(receptor_cif, factura_details, emisor_cif):
    emisor = cfdi40.Emisor(
        rfc=emisor_cif['Rfc'],
        nombre=emisor_cif['RazonSocial'],
        regimen_fiscal=emisor_cif['RegimenFiscal']
    )
    emisor = emisor | factura_details.get('Emisor', {})
    invoice = cfdi40.Comprobante(
        emisor=emisor,
        lugar_expedicion=emisor_cif['CodigoPostal'],
        receptor=cfdi40.Receptor(
            rfc=factura_details['Receptor'],
            nombre=receptor_cif['RazonSocial'],
            uso_cfdi=factura_details['UsoCFDI'],
            domicilio_fiscal_receptor=receptor_cif['CodigoPostal'],
            regimen_fiscal_receptor=receptor_cif['RegimenFiscal']
        ),
        metodo_pago=factura_details['MetodoPago'],
        forma_pago=factura_details['FormaPago'],
        # serie=serie,
        # folio=folio,
        conceptos=factura_details["Conceptos"]
    )
    invoice = invoice.process()

    expected_total = factura_details.get('Total')
    if expected_total is not None:
        if expected_total != invoice['Total']:
            logger.info(f"{factura_details['Receptor']}: Total '{expected_total}' is invalid, expected '{invoice['Total']}'")
            return None

    return invoice


def parse_periodo_mes_ajuste(periodo_mes_ajuste: str):
    parts = periodo_mes_ajuste.split(".")
    if len(parts) != 2:
        raise ValueError("Periodo Invalido")

    periodo, mes_ajuste = parts
    if not mes_ajuste.isnumeric():
        raise ValueError("Periodo Invalido")

    mes_ajuste = int(mes_ajuste)
    if not (12 >= int(mes_ajuste) >= 1):
        raise ValueError("Periodo Invalido")

    if periodo not in PERIODICIDAD:
        raise ValueError("Periodo Invalido")

    return PERIODICIDAD[periodo], mes_ajuste


def format_concepto_desc(concepto, periodo):
    concepto = concepto.copy()
    template = facturacion_environment.from_string(concepto["Descripcion"])
    concepto["Descripcion"] = template.render(
        periodo=periodo
    )
    return concepto


def validad_facturas(clients, facturas):
    is_valid = True
    for factura_details in facturas:
        cliente = clients.get(factura_details['Receptor'])
        if not cliente:
            logger.info(f"{factura_details['Receptor']}: client not found")
            is_valid = False

        for c in factura_details["Conceptos"]:
            periodo_mes_ajuste = c.get("_periodo_mes_ajuste", "")
            try:
                parse_periodo_mes_ajuste(periodo_mes_ajuste)
            except ValueError:
                logger.info(f"{factura_details['Receptor']}: _periodo_mes_ajuste '{periodo_mes_ajuste}' is invalid")
                is_valid = False

        if factura_details["MetodoPago"] == "PPD" and factura_details["FormaPago"] != "99":
            logger.info(f"{factura_details['Receptor']}: FormaPago '{factura_details['FormaPago']}' is invalid, expected '99' for PPD")
            is_valid = False

    return is_valid


def period_desc(dp: DatePeriod):
    return format_date(date(year=dp.year, month=dp.month, day=1), locale='es_MX', format="'Mes de' MMMM 'del' y").upper()


def periodicidad_desc(dp: DatePeriod, periodo_mes_ajuste, offset):
    periodo_meses, mes_ajuste = parse_periodo_mes_ajuste(periodo_mes_ajuste)

    if (dp.month - mes_ajuste) % periodo_meses == 0:
        if offset:
            dp = add_month(dp, offset)

        periodo = period_desc(dp)
        if periodo_meses > 1:
            periodo += " AL " + period_desc(
                dp=add_month(dp, periodo_meses - 1)
            )

        return periodo
    return None


def generate_ingresos(clients, facturas, dp, emisor_rfc):
    if not validad_facturas(clients, facturas):
        return

    emisor_cif = clients[emisor_rfc]

    def prepare_concepto(concepto):
        periodo = periodicidad_desc(
            dp,
            concepto['_periodo_mes_ajuste'],
            concepto.get('_desfase_mes')
        )
        if periodo and concepto['ValorUnitario'] is not None:
            return format_concepto_desc(concepto, periodo=periodo)

    def facturas_iter():
        i = 0
        for f in facturas:
            receptor_cif = clients[f['Receptor']]
            conceptos = [x for x in (prepare_concepto(c) for c in f["Conceptos"]) if x]
            if conceptos:
                f["Conceptos"] = conceptos
                yield create_cfdi(receptor_cif, f, emisor_cif)
                i += 1

    cfdis = [i for i in facturas_iter()]

    if None in cfdis:
        return

    return cfdis


def parse_fecha_pago(fecha_pago):
    if not fecha_pago:
        raise ValueError("Fecha de Pago es requerida")

    fecha_pago = datetime.fromisoformat(fecha_pago)
    if fecha_pago > datetime.now():
        raise ValueError("Fecha de Pago es mayor a la fecha actual")

    if fecha_pago.replace(hour=12) > datetime.now():
        fecha_pago = datetime.now()
    else:
        fecha_pago = fecha_pago.replace(hour=12)

    dif = datetime.now() - fecha_pago
    if dif.days > 30:
        raise ValueError("Fecha de Pago es de hace mas de 30 dias")

    return fecha_pago


def parse_importe_pago(importe_pago: str):
    if not importe_pago:
        return None

    try:
        return round(Decimal(importe_pago), 2)
    except InvalidOperation:
        raise ValueError("Importe de Pago es invalido")


def pago_factura(factura_pagar, fecha_pago: datetime, forma_pago: str, importe_pago: Decimal = None):
    c = factura_pagar
    invoice = Comprobante.pago_comprobantes(
        comprobantes=[
            PagoComprobante(
                comprobante=c,
                num_parcialidad=c.ultima_num_parcialidad + 1,
                imp_saldo_ant=c.saldo_pendiente,
                imp_pagado=importe_pago
            )
        ],
        fecha_pago=fecha_pago,
        forma_pago=forma_pago,
    )
    return invoice.process()


def find_ajustes(facturas, mes_ajuste):
    for f in facturas:
        rfc = f["Receptor"]
        for concepto in f["Conceptos"]:
            _, mes_aj = parse_periodo_mes_ajuste(concepto['_periodo_mes_ajuste'])
            if mes_aj == mes_ajuste:
                yield rfc, concepto


def archivos_folder(dp: DatePeriod):
    if dp.month:
        return os.path.join(ARCHIVOS_DIRECTORY, str(dp.year), str(dp.year) + "-{:02d}".format(dp.month))
    return os.path.join(ARCHIVOS_DIRECTORY, str(dp.year))


def archivos_filename(dp: DatePeriod, ext="xlsx"):
    return os.path.join(archivos_folder(dp), f"{dp}.{ext}")


def exportar_facturas(all_invoices, dp: DatePeriod, emisor_cif, rfc_prediales):
    emisor_rfc = emisor_cif['Rfc']
    emisor_regimen = emisor_cif['RegimenFiscal']

    emitidas = filter_invoices_iter(invoices=all_invoices.values(), fecha=dp, rfc_emisor=emisor_rfc)
    emitidas_pagos = filter_payments_iter(invoices=all_invoices, fecha=dp, rfc_emisor=emisor_rfc)
    emitidas_pagos = list(emitidas_pagos)

    recibidas = filter_invoices_iter(invoices=all_invoices.values(), fecha=dp, rfc_receptor=emisor_rfc)
    recibidas_pagos = filter_payments_iter(invoices=all_invoices, fecha=dp, rfc_receptor=emisor_rfc)

    recibidas_pagos = list(recibidas_pagos)
    pagos_hechos_iva = [
        p
        for p in recibidas_pagos
        if sum(x.get("Importe", 0) for x in p.impuestos.get("Traslados", {}).values()) > 0
           and p.comprobante["Receptor"].get("RegimenFiscalReceptor") in (emisor_regimen, None)
    ]
    prediales = [p for p in recibidas_pagos if p.comprobante["Emisor"]["Rfc"] in rfc_prediales]

    archivo_excel = archivos_filename(dp)
    os.makedirs(os.path.dirname(archivo_excel), exist_ok=True)

    workbook = xlsxwriter.Workbook(archivo_excel)
    # EMITIDAS
    invoices_export(workbook, "EMITIDAS", emitidas)
    payments_export(workbook, "EMITIDAS PAGOS", emitidas_pagos)

    # RECIBIDAS
    invoices_export(workbook, "RECIBIDAS", recibidas)
    payments_export(workbook, "RECIBIDAS PAGOS", recibidas_pagos)

    # SPECIALES
    payments_export(workbook, f"RECIBIDAS PAGOS IVA {emisor_regimen.code}", pagos_hechos_iva)
    if prediales:
        payments_export(workbook, "PREDIALES", prediales)

    # RETENCIONES
    if dp.month is None:
        archivo_retenciones = archivos_filename(dp, ext="retenciones.txt")
        pagos_agrupados = payments_groupby_receptor(emitidas_pagos)
        payments_retentions_export(archivo_retenciones, pagos_agrupados)

    try:
        workbook.close()
        return archivo_excel
    except FileCreateError:
        return None


def generate_pdf_template(template_name, fields):
    increment_template = facturacion_environment.get_template(template_name)
    md5_document = increment_template.render(
        fields
    )
    html = markdown(md5_document)
    pdf = HTML(string=html).write_pdf(
        target=None,
        stylesheets=[
            os.path.join(SOURCE_DIRECTORY, "markdown_styles", "markdown6.css"),
            CSS(
                string='@page { width: Letter; margin: 1.6cm 1.6cm 1.6cm 1.6cm; }'
            )
        ]
    )
    return pdf


def mf_pago_fmt(cfdi):
    i = cfdi
    if i['TipoDeComprobante'] == "I":
        return i['TipoDeComprobante'].code + ' ' + i['MetodoPago'].code + ' ' + (i['FormaPago'].code if i['FormaPago'].code != '99' else '')
    return i['TipoDeComprobante'].code


def ajustes_directory(dp: DatePeriod):
    return os.path.join(archivos_folder(dp), 'ajustes')


def preview_cfdis(cfdis):
    outfile = os.path.join(TEMP_DIRECTORY, "factura.html")
    Representable.html_write_all(
        objs=cfdis,
        target=outfile,
    )
    os.startfile(
        os.path.abspath(outfile)
    )



from datetime import date, datetime

import pytest
from satcfdi import DatePeriod

from satdigitalinvoice.facturacion import FacturacionGUI
from yaml.constructor import ConstructorError

from satdigitalinvoice.file_data_managers import LocalData, ConfigManager, FacturasManager

from satdigitalinvoice.layout import make_layout
from satdigitalinvoice.utils import random_string, add_month


def test_layout_unique_keys():
    try:
        import _tkinter
    except ImportError:
        return

    layout = make_layout(has_fiel=True)

    def elements(layout):
        for e in layout:
            if isinstance(e, list):
                yield from elements(e)
            else:
                yield e

    unique_keys = set()
    for e in elements(layout):
        if e.Key:
            if e.Key in unique_keys:
                raise Exception(f"Key {e.Key} is not unique")
            else:
                unique_keys.add(e.Key)


def test_duplicated_yaml_file():
    class MyConfigManager(ConfigManager):
        file_source = "config_duplicated.yaml"

    # expect exception thrown
    with pytest.raises(ConstructorError) as e:
        MyConfigManager()
    assert e.value.problem == 'found duplicate key (email_notificada_hasta)'


def test_random_string():
    a = random_string()
    assert len(a) == 32


def test_duplicated_facturas():
    class MyFacturasManager(FacturasManager):
        file_source = "facturas_duplicated.yaml"

    # expect exception thrown

    with pytest.raises(ValueError) as e:
        MyFacturasManager(DatePeriod(2021, 1, 1))
    assert str(e.value)[0:20] == 'Factura Duplicada: {'


def test_increase_month():
    d = DatePeriod(2021, 1)
    d = add_month(d, 1)
    assert d.year == 2021 and d.month == 2


def test_app_setup():
    config = ConfigManager()
    a = FacturacionGUI(
        config
    )



from datetime import date, datetime

import PySimpleGUI as sg
# noinspection PyUnresolvedReferences
from satcfdi.transform.catalog import CATALOGS

from .log_tools import *

FORMA_PAGO = CATALOGS['{http://www.sat.gob.mx/sitio_internet/cfd/catalogos}c_FormaPago']
TEXT_PADDING = ((5, 0), 3)

# 24 x 24
FOLDER_ICON = "iVBORw0KGgoAAAANSUhEUgAAABgAAAAYCAYAAADgdz34AAAACXBIWXMAAAsTAAALEwEAmpwYAAAAqElEQVR4nO3UsQnCYBRF4U/EAcRJtHUJcQddwQlsxcpVLF0hOoY2FlaC+SUQQWI0PmwscuF0j3PhFZc2" \
              "/5QR9siRnjhhid6vBVlFXGWHwS8Ft4aCHLPydozFGybo1BV8kl8wLe/muDbcryIFBwzRxbpB/OAcfV8f2y/lD17y7qcbHIPy2oKoILUFqX1RdCpSgMIVHrsUoHDVznVWM9cpQF46imlp8ye5AyE7C1To4" \
              "/HLAAAAAElFTkSuQmCC"

EXCEL_ICON = "iVBORw0KGgoAAAANSUhEUgAAABgAAAAYCAYAAADgdz34AAAACXBIWXMAAAsTAAALEwEAmpwYAAAA7klEQVR4nN2VQQqCUBCGv2pfFyhv0FmC7hGFq1bhrgi6gi1cBK6CdhFtvYdCi6CgrZsXwgTyUNOnLmrgBx31" \
             "/+DNOAP/HkNgCqyBK3AHLsDA1HAELIEjcANUjuYm5hbwKDBVKTnyjZ5/ATugkwWwS5qrAsBHkyyA0yBg1SQgAraiSHueCwgBL3V" \
             "/kFwWwAd6Ir8swJXcSZTEvsIRfQXEwFg6y5LruElAogDoioKma1AF4JvUoPUjctsucqi1qVfQpsY1KPuj+SY1qDsqnLaH3SJvF9Qd1wp4ilethTOT9zea7CLzsivzDPSrmPx2vAFRfA9plcmcVwAAAABJRU5ErkJggg=="

HTML_ICON = "iVBORw0KGgoAAAANSUhEUgAAABgAAAAYCAYAAADgdz34AAAACXBIWXMAAAsTAAALEwEAmpwYAAAA/ElEQVR4nN2VOwoCMRCGPz2HB7ARHyfwBN5EsLOxWKzs7Ox0e69gYbOtoK2dZ7BVRwZGWJbsJvuw0IGfDUkmX" \
            "/gzm8A/xgi4AC9AHNLo1gGccxaWFCAC1lUBz0CAVIVICYAYpPVNgJSF+ABtYOHoXzcF6AMdYAmsMuo1ATgZpJTvZQAScEa5ETcAiIsA+vcmDm9XgUpsjUJ7tASnGV1tTL8z26VeKTdgAOwz5esFqA6mKDX2WUR3ebe" \
            "+MXCsAkhMLsAGmNQBZC3C7MHRpxY9rO0FvMyWeUUdfIe8a6BMt9S8rqVAmlv7wZECaa43RjYx78kUh3Su5gxDAL8Vb7wE7yjidnvCAAAAAElFTkSuQmCC"

PDF_ICON = "iVBORw0KGgoAAAANSUhEUgAAABgAAAAYCAYAAADgdz34AAAACXBIWXMAAAsTAAALEwEAmpwYAAAA+0lEQVR4nN3VPUrEQBQA4E89gCBio7WI" \
           "+INbeAbxAoI3ESzEysoFd21tbe28QixE7a2t7dxGRwKvWCTrTpJNoQ9ek5k3X8hM5vEfo4cXfCFVZBkbbYDnCQunMeAM" \
           "/abAZyaQmiKpBpACmesSSHWRacA8Tiue92cF7GIN57j4kduzAB4DqfXdc4ACQ3xk7FEjYBk7GHQFrOMQl10BJ3GC7roC3mL8vivgNsYXsYmlyKIt8IojLOAYV1jFQWz6sA1QxOnZwv" \
           "7YFf6OG1xj1AYYRH9YwVPGT5jqXtejeMuHzMXL2tYNJ/2SZe3U6MXESS0zVWQ5t6zZywH+VnwDCQv1frFQIlYAAAAASUVORK5CYII="

ABOUT_ICON = "iVBORw0KGgoAAAANSUhEUgAAABgAAAAYCAYAAADgdz34AAAACXBIWXMAAAsTAAALEwEAmpwYAAABP0lEQVR4nOWVTU7DMBCFvxuwKmrX0G4o5QLtQboDFXEbwrLkGNDucwB" \
             "+VvQObS5ApCxAIz1LCNmOG68QTxopst882y/jMfwHnAIroAQ2ilJjgxzhkYRa4AvYA6" \
             "+KvcZacYx7FBZADXwC98DUw7GxAmjEXRwjbkkfwHkCfyyu5cy7yCPtxhJOUnckruUcgGGMWMqWs8D8UhE6iZ3iMVYtrTwPoVKEUEjDW123qowL" \
             "+uNSGjche6z8YlhGLHI4hGzaAi8dyVWHReiePPsmNprMXeAdeOprUZWwQB2yaKUfNM1Y4Eoa17EyLTIWeIiVqbOp0aXx4U7hw0S565RWsevRKnYpreJ3sxsniE8kntTsHObaTaN" \
             "/MvNwZvK8ETdZ3GGocnMPjom8KeofD846xZYYBuotJmQ31MK+bSzryfwb+AbymF7gpXVM1QAAAABJRU5ErkJggg=="

REFRESH_ICON = "iVBORw0KGgoAAAANSUhEUgAAABgAAAAYCAYAAADgdz34AAAACXBIWXMAAAsTAAALEwEAmpwYAAABGUlEQVR4nO3VzSqFURTG8Z8JYmRiJN9yDQYiKbdhYqDO9OQWyMSQDGRISoYMlHIJJEWZ" \
               "+v4oysTRrqU4nePdZej91+ptt3frWWu31/NS8kdaMINNXOANLzjFGia/ne3CdHyzGMUxarjBNpawgt0QSnsHGA6xWp1oU8bxhGvMorXBmQ5UQugO1VyBITziBD0ZxQziCh+5Aoe4z0y+ELEeyQsF0r2" \
               "/Y14etQZR2EE32jIFSv4Z7fFAfuVr3Osjh0o88ZFcgS0sRhTRF9OffEmOQBr7S/RnJO/FWUz" \
               "/QK5ANQzsOVrvbHA2DeMcbvGAsYxiflhuMr39WCfH3AurXsZOFJD2joruXcFPYwKr4a6po1ecYwNTuYlLNOMTqXpPNQBv/ywAAAAASUVORK5CYII="

PREVIEW_ICON = "iVBORw0KGgoAAAANSUhEUgAAABgAAAAYCAYAAADgdz34AAAACXBIWXMAAAsTAAALEwEAmpwYAAABKElEQVR4nO2VMU8CQRCFP6WAQksqOxJ/A6WJUFFprJSEilARWgsa" \
               "/4IVLYUNFZ0WVECwNxpCIDQYfwEVGj0zycNcCOTulqMxvGQze2/eztub3O3Cf0QJ6IQYD8CRi8Ed4AUUn0rTczFZGoTRfLmYRDG4Aj6BAXC8C4M80NC8H/ZNohh4K6Mel0FGu" \
               "/cPT2tjMViHvUEg9i0KxBmQjeszPVSxCnCrmBW/lUECqAKzDX/yTPmEi0EaeFauq0JDzU+BG809HX7pKAYnwAiYA9fi7qVr+jZgKEo30rqNBh2dlAe65Yx7BJLStX2teQFa4i3/JN7WBV6Z5+JrwA/wKu4S" \
               "+FYhixdADniTzvSRUQDeVXSsNiwUJ+ItbzpnpICy2vAhA4v2bLzl//ALOtSCJzC7jH4AAAAASUVORK5CYII="

SEARCH_ICON = "iVBORw0KGgoAAAANSUhEUgAAABgAAAAYCAYAAADgdz34AAAACXBIWXMAAAsTAAALEwEAmpwYAAABM0lEQVR4nOXTuytAYRzG8Y8RJTPJ5jJIitwGoxiZGSTZDAaDTQaJhYWBf0A" \
              "WRDHjD7BIMrkUg2wol079lEGc1yXJU2c4b8/5Pu/vfZ/Df1MjJrGKLSyjHyVfBZdjE0+4wwH2cB5rVxj8LLwmQNcYeWO3TdiOoJlUeDGOcIbqd3wFmI2QgZSAcTyiPYc3C9nBZcq" \
              "dnGAjYUPNMUVfHnNlmIcSArIpLrCUx9wSAZ3StB8V/lD1EdCTGJBVeCVvg+4wlQAvxT0m8n6wERUtyukfjakb8ga04QELOby1uMGaxFbsxq4W49jeUke05xRlKfD5gO/GJBlkGr3" \
              "ownA0JvsZD1H1Gfh8vLdiHbex/vIcYwyFX4G/VnbhdRFYIVEZbO7VmRekAn4Nnqn7nWP5NnX9JNyf1DM+Yku7BVpvYgAAAABJRU5ErkJggg=="

CONFIG_ICON = "iVBORw0KGgoAAAANSUhEUgAAABgAAAAYCAYAAADgdz34AAAACXBIWXMAAAsTAAALEwEAmpwYAAABPElEQVR4nOWVvWoCQRSFP5NCSJHOdGkECaa3s" \
              "/MtUu6CYOcTaGMZU2efwNYtrKwkWPgK9nmAEC38IRMGTmCZrIMru6B44MLOmXPP7N6ZvQPXgiowdcJyueEFMMBMYcTlhjfgG7hR2OfhqWYB0AEqQBnoA1/APKGZi" \
              "+tLU1GOzfXiEVirBDvgE9gA70A9oauL20izU85aHgcxAlZACxgAH0DDo29IM1DOSh6paAI/QJQyVwJCIFaE4lxE8rBe//AALIA90HbMY5VgqTDikou0lbuQVyrugIkMnsSFGttN/ENHnJ1DWqNc6+FFV" \
              "+J7jWO9tYul5pDWKJc8FxhnXSBriQKNa8eUqPBNbh5xTMeK4JRjWviP5msVkdMqnsVts7aKtGbX8zS7XtZmdwhDff5tol2/ckkXTrXoK/N88Qvnr38CSEQRlwAAAABJRU5ErkJggg=="

EDIT_ICON = "iVBORw0KGgoAAAANSUhEUgAAABgAAAAYCAYAAADgdz34AAAACXBIWXMAAAsTAAALEwEAmpwYAAAA10lEQVR4nO3UPUpDQRSG4cfCQgn" \
            "+LSPLSCcEC1u3EXdgEXdg3EC2IXYS0N4itZWWGkGvDHyBkE45Npq3usyB95tz5s6w4ZfZwwQveMIFtqvkB5jhDZe4xmdCyuQdxivrk3RSIl/gAe84Ta118oytCvkQPdwmZJxxXVXJl/TSSZd6O/hvs4+7yE" \
            "/Wasd4xT2ONvIl/3gsuys3dLRWGyZ0ll/2RwwiX2SnZ5XyxnkerD5uEjKqkjemmOe7H3FXJW885rGaR9xVyg/xkZBpxjXIwZewU7VTf4ovf6VSMafchm4AAAAASUVORK5CYII="

BUTTON_COLOR = (sg.theme_background_color(), sg.theme_background_color())


def make_layout(has_fiel):
    # LAYOUT
    button_column = [
        sg.Button(image_data=CONFIG_ICON, key="ver_config", border_width=0, button_color=BUTTON_COLOR),
        sg.Text("Periodo:", pad=TEXT_PADDING),
        sg.Input(date.today().strftime('%Y-%m'), size=(11, 1), key="periodo"),
        sg.Button(image_data=FOLDER_ICON, key="ver_carpeta", border_width=0, button_color=BUTTON_COLOR),
        sg.Button(image_data=EXCEL_ICON, key="ver_excel", border_width=0, button_color=BUTTON_COLOR),
        sg.Button(image_data=HTML_ICON, key="ver_html", border_width=0, button_color=BUTTON_COLOR),

        sg.Push(),
        sg.Text("Factura:", pad=TEXT_PADDING),
        sg.Text("", key="serie", pad=TEXT_PADDING, text_color="black"),
        sg.Input("", key="folio", size=(8, 1), enable_events=True),
        sg.Button("".center(22), disabled=True, key="crear_facturas", border_width=0, button_color=sg.theme_background_color()),
    ]

    # ----- Full layout -----
    return [
        button_column,
        [
            sg.TabGroup(
                [[
                    sg.Tab(
                        'Consola'.center(13),
                        [
                            [
                                sg.Push(),
                                sg.Button(image_data=ABOUT_ICON, key="about", border_width=0, button_color=BUTTON_COLOR),
                            ],
                            [sg.Multiline(
                                expand_x=True,
                                expand_y=True,
                                key="console",
                                write_only=True,
                                autoscroll=True,
                                reroute_stdout=True
                            )]
                        ],
                        key='console_tab',
                    ),
                    sg.Tab(
                        'Clientes'.center(13),
                        [
                            [
                                sg.Button("Refrescar", key="refresh_clientes", border_width=0),
                                sg.Push(),
                                sg.Button(image_data=EDIT_ICON, key="editar_clientes", border_width=0, button_color=BUTTON_COLOR),
                            ],
                            [
                                sg.Table(
                                    values=[],
                                    headings=[
                                        "#",
                                        "Razon Social",
                                        "Rfc",
                                        "Reg",
                                        "CP",
                                        "Facturas",
                                    ],
                                    key="clientes_table",
                                    expand_x=True,
                                    expand_y=True,
                                    select_mode=sg.TABLE_SELECT_MODE_EXTENDED,
                                    enable_events=True,
                                    text_color="black",
                                    background_color="white",
                                    def_col_width=10,
                                )
                            ]],
                        key='clients_tab',
                    ),
                    sg.Tab(
                        'Facturas'.center(13),
                        [
                            [
                                sg.Button("Refrescar", key="refresh_facturas", border_width=0, ),
                                sg.Text("", pad=TEXT_PADDING, key="preparar_facturas_text"),
                                sg.Push(),
                                sg.Button(image_data=EDIT_ICON, key="editar_facturas", border_width=0, button_color=BUTTON_COLOR),
                            ],
                            [
                                sg.Table(
                                    values=[],
                                    headings=[
                                        '#',
                                        'EReg',
                                        'Receptor Razon Social',
                                        'Recep. Rfc',
                                        "Tipo",
                                        "Subtotal",
                                        "Total"
                                    ],
                                    key="facturas_table",
                                    expand_x=True,
                                    expand_y=True,
                                    select_mode=sg.TABLE_SELECT_MODE_EXTENDED,
                                    enable_events=True,
                                    text_color="black",
                                    background_color="white",
                                )
                            ]],
                        key='facturas_tab',
                    ),
                    sg.Tab(
                        'Emitidas'.center(13),
                        [
                            [
                                sg.Button("Pendientes", key="facturas_pendientes", border_width=0),
                                sg.Button("Todas", key="facturas_emitidas", border_width=0),
                                sg.Button(image_data=SEARCH_ICON, key="emitidas_search_enter", border_width=0, button_color=BUTTON_COLOR),
                                sg.Input("", size=(20, 1), key="emitidas_search", border_width=0),
                                sg.Text("", pad=TEXT_PADDING, key="emitidas_text"),
                            ],
                            [
                                sg.Column([[
                                    sg.Button("".ljust(10), key="status_sat", border_width=0, button_color=sg.theme_background_color()),
                                    sg.Button("".ljust(10), key="email_notificada", border_width=0, button_color=sg.theme_background_color()),
                                    sg.Button("".ljust(10), key="pendiente_pago", border_width=0, button_color=sg.theme_background_color()),
                                ]]),
                                sg.VSeparator(),
                                sg.Column([[
                                    sg.CalendarButton("FechaPago:", format='%Y-%m-%d', title="FechaPago", no_titlebar=False, target="fecha_pago", pad=TEXT_PADDING,
                                                      border_width=0,
                                                      key="fecha_pago_select"),
                                    sg.Input(datetime.now().strftime('%Y-%m-%d'), size=(12, 1), key="fecha_pago", border_width=0),
                                    sg.Combo([Code(k, v) for k, v in FORMA_PAGO.items()], default_value=Code("03", FORMA_PAGO["03"]), key="forma_pago", size=(28, 1)),
                                    sg.Text("ImpPagado:", pad=TEXT_PADDING, key="imp_pagado_text", border_width=0),
                                    sg.Input("", size=(12, 1), key="importe_pago", border_width=0),
                                    sg.Button("Comprobante Pago", key="prepare_pago", border_width=0),
                                    sg.Button(image_data=PREVIEW_ICON, key="ver_html_pago", border_width=0, button_color=BUTTON_COLOR),
                                ]], visible=False, key="ppd_action_items"),
                            ],
                            [
                                sg.Table(
                                    values=[],
                                    headings=[
                                        '#',
                                        'Receptor Razon Social',
                                        'Recep. Rfc',
                                        'Factura',
                                        "Fecha",
                                        "Total",
                                        "Pagada",
                                        "Tipo",
                                    ],
                                    key="emitidas_table",
                                    expand_x=True,
                                    expand_y=True,
                                    select_mode=sg.TABLE_SELECT_MODE_BROWSE,
                                    enable_events=True,
                                    text_color="black",
                                    background_color="white",
                                )
                            ]],
                        key='emitidas_tab',

                    ),
                    sg.Tab(
                        'Correos'.center(13),
                        [
                            [
                                sg.Button("Refrescar", key="refresh_correos", border_width=0, ),
                            ],
                            [
                                sg.Table(
                                    values=[],
                                    headings=[
                                        '#',
                                        'Receptor Razon Social',
                                        'Recep. Rfc',
                                        'Facturas',
                                        'Pendientes Emitidas Meses Anteriores'
                                    ],
                                    key="correos_table",
                                    expand_x=True,
                                    expand_y=True,
                                    select_mode=sg.TABLE_SELECT_MODE_EXTENDED,
                                    enable_events=True,
                                    text_color="black",
                                    background_color="white",
                                    def_col_width=10,
                                )
                            ]],
                        key='correos_tab',
                    ),
                    sg.Tab(
                        'Ajustes'.center(13),
                        [
                            [
                                sg.Button("Refrescar", key="refresh_ajustes", border_width=0, ),
                                sg.Text("", pad=TEXT_PADDING, key="preparar_ajustes_text"),
                            ],
                            [
                                sg.Table(
                                    values=[],
                                    headings=[
                                        "#",
                                        "Receptor Razon Social",
                                        "Recep. Rfc",
                                        "Actual",
                                        "Nuevo",
                                        "Ajuste %",
                                        "Periodo",
                                        "Ajuste Periodo",
                                        "Ajuste Efectivo"
                                    ],
                                    key="ajustes_table",
                                    expand_x=True,
                                    expand_y=True,
                                    right_click_selects=True,
                                    select_mode=sg.TABLE_SELECT_MODE_EXTENDED,
                                    enable_events=True,
                                    text_color="black",
                                    background_color="white",
                                    def_col_width=10,
                                )
                            ]],
                        key='ajustes_tab'
                    ),
                    sg.Tab(
                        'Recuperar'.center(13),
                        [
                            [
                                sg.Button("SAT Status", key="sat_status_todas", border_width=0),
                                sg.Text("Recuperar:", pad=TEXT_PADDING),
                                sg.Button("Emitidas ", key="recuperar_emitidas", border_width=0),
                                sg.Button("Recibidas", key="recuperar_recibidas", border_width=0),
                                sg.Text("Dias:", pad=TEXT_PADDING),
                                sg.Input("40", size=(4, 1), key="recuperar_dias"),
                            ]
                        ],
                        key='recuperar_tab',
                        visible=has_fiel
                    ),
                ]],
                expand_x=True,
                expand_y=True,
                enable_events=True,
                key="main_tab_group",
            )
        ]
    ]


class ActionButtonManager:
    def __init__(self, button):
        self.name = ""
        self.items = []
        self.button = button

    def set_items(self, name, items):
        self.name = name
        self.items = items
        self.style_button()

    def clear(self):
        self.name = ""
        self.items = []
        self.style_button()

    def text(self):
        if self.items:
            return f"Procesar {len(self.items)} {self.name.capitalize()}"
        else:
            return ""

    def style_button(self):
        self.button.update(
            self.text().center(22),
            disabled=not self.items,
            button_color=sg.theme_button_color() if self.items else sg.theme_background_color()
        )



import collections.abc
import json
import logging
import os
from decimal import Decimal

import jsonschema as jsonschema
import yaml
from satcfdi import Code, DatePeriod
from satcfdi.pacs import sat
# noinspection PyUnresolvedReferences
from satcfdi.transform.catalog import CATALOGS
from satcfdi.transform.helpers import Xint
from yaml import MappingNode, SafeLoader
from yaml.constructor import ConstructorError

from . import SOURCE_DIRECTORY
from .utils import find_best_match, first_duplicate

REGIMEN_FISCAL = CATALOGS['{http://www.sat.gob.mx/sitio_internet/cfd/catalogos}c_RegimenFiscal']

logger = logging.getLogger(__name__)
sat_manager = sat.SAT()


class DuplicateKeySafeLoader(yaml.CSafeLoader):
    def construct_mapping(self, node, deep=False):
        if isinstance(node, MappingNode):
            self.flatten_mapping(node)

        if not isinstance(node, MappingNode):
            raise ConstructorError(None, None,
                                   "expected a mapping node, but found %s" % node.id,
                                   node.start_mark)
        mapping = {}
        for key_node, value_node in node.value:
            key = self.construct_object(key_node, deep=deep)
            if not isinstance(key, collections.abc.Hashable):
                raise ConstructorError("while constructing a mapping", node.start_mark,
                                       "found unhashable key", key_node.start_mark)
            value = self.construct_object(value_node, deep=deep)
            if key in mapping:
                raise ConstructorError("while constructing a mapping", node.start_mark,
                                       "found duplicate key (%s)" % key, key_node.start_mark)
            mapping[key] = value
        return mapping


def load_validator(schema_file):
    with open(os.path.join(SOURCE_DIRECTORY, 'schemas', schema_file), "r", encoding="utf-8") as fs:
        schema = yaml.load(fs, SafeLoader)
        validator = jsonschema.validators.validator_for(schema)
        validator.check_schema(schema)
        return validator(schema)


client_validator = load_validator("cliente.yaml")
factura_validator = load_validator("factura.yaml")
config_validator = load_validator("config.yaml")


class LocalData(dict):
    file_source = None

    def __new__(cls, *args, **kwargs):
        return super().__new__(cls)

    def __init__(self):
        super().__init__(self._raw())

    def _raw(self):
        with open(self.file_source, "r", encoding="utf-8") as fs:
            return yaml.load(fs, DuplicateKeySafeLoader)


class ConfigManager(LocalData):
    file_source = "config.yaml"

    def __init__(self):
        super().__init__()
        if error := jsonschema.exceptions.best_match(config_validator.iter_errors(self)):
            raise error


class ClientsManager(LocalData):
    file_source = "clientes.yaml"

    def __init__(self):
        super().__init__()
        for k, v in self.items():
            if error := jsonschema.exceptions.best_match(client_validator.iter_errors(v)):
                raise error
            self[k]["Rfc"] = k
            self[k]["RegimenFiscal"] = Code(self[k]["RegimenFiscal"], REGIMEN_FISCAL.get(self[k]["RegimenFiscal"]))


class FacturasManager(LocalData):
    file_source = "facturas.yaml"

    def __init__(self, dp: DatePeriod | None):
        def loading_function(loader, node):
            cases = loader.construct_mapping(node, deep=True)
            if dp is None:
                return cases
            return find_best_match(
                cases,
                dp
            )[1]

        DuplicateKeySafeLoader.add_constructor("!case", loading_function)
        super().__init__()
        if dup := first_duplicate(json.dumps(x, sort_keys=True, default=str) for x in self["Facturas"]):
            raise ValueError("Factura Duplicada: {}".format(dup))

        if dp:
            for v in self["Facturas"]:
                if error := jsonschema.exceptions.best_match(factura_validator.iter_errors(v)):
                    raise error


DuplicateKeySafeLoader.add_constructor("!decimal", lambda loader, node: Decimal(loader.construct_scalar(node)))
DuplicateKeySafeLoader.add_constructor("!read", lambda loader, node: open(loader.construct_scalar(node), 'rb').read())


def represent_decimal(dumper, data):
    return dumper.represent_scalar('tag:yaml.org,2002:str', str(data))


def represent_str(dumper, data):
    return dumper.represent_scalar('tag:yaml.org,2002:str', str(data))


yaml.SafeDumper.add_multi_representer(dict, lambda dumper, data: dumper.represent_dict(data))
yaml.SafeDumper.add_representer(Decimal, represent_decimal)
yaml.SafeDumper.add_representer(Code, represent_str)
yaml.SafeDumper.add_representer(Xint, represent_str)
